"""
This module defines handlers for storing sessions when handles
sessions of users connecting to the server.

There are two similar but separate stores of sessions:

- ServerSessionHandler - this stores generic game sessions
     for the game. These sessions has no knowledge about
     how they are connected to the world.
- PortalSessionHandler - this stores sessions created by
     twisted protocols. These are dumb connectors that
     handle network communication but holds no game info.

"""
import time
from codecs import decode as codecs_decode

from django.conf import settings
from django.utils.translation import gettext as _

import evennia
from evennia.commands.cmdhandler import CMD_LOGINSTART
from evennia.server.portal import amp
from evennia.server.signals import (
    SIGNAL_ACCOUNT_POST_FIRST_LOGIN,
    SIGNAL_ACCOUNT_POST_LAST_LOGOUT,
    SIGNAL_ACCOUNT_POST_LOGIN,
    SIGNAL_ACCOUNT_POST_LOGOUT,
)
from evennia.utils.logger import log_trace
from evennia.utils.utils import (
    callables_from_module,
    class_from_module,
    delay,
    is_iter,
    make_iter,
)

_FUNCPARSER_PARSE_OUTGOING_MESSAGES_ENABLED = settings.FUNCPARSER_PARSE_OUTGOING_MESSAGES_ENABLED
_BROADCAST_SERVER_RESTART_MESSAGES = settings.BROADCAST_SERVER_RESTART_MESSAGES

# delayed imports
_AccountDB = None
_ServerSession = None
_ServerConfig = None
_ScriptDB = None
_OOB_HANDLER = None

_ERR_BAD_UTF8 = _("Your client sent an incorrect UTF-8 sequence.")


class DummySession(object):
    sessid = 0


DUMMYSESSION = DummySession()

_SERVERNAME = settings.SERVERNAME
_MULTISESSION_MODE = settings.MULTISESSION_MODE
_IDLE_TIMEOUT = settings.IDLE_TIMEOUT
_DELAY_CMD_LOGINSTART = settings.DELAY_CMD_LOGINSTART
_MAX_SERVER_COMMANDS_PER_SECOND = 100.0
_MAX_SESSION_COMMANDS_PER_SECOND = 5.0
_MODEL_MAP = None
_FUNCPARSER = None


# input handlers

_INPUT_FUNCS = {}
for modname in make_iter(settings.INPUT_FUNC_MODULES):
    _INPUT_FUNCS.update(callables_from_module(modname))


def delayed_import():
    """
    Helper method for delayed import of all needed entities.

    """
    global _ServerSession, _AccountDB, _ServerConfig, _ScriptDB
    if not _ServerSession:
        # we allow optional arbitrary serversession class for overloading
        _ServerSession = class_from_module(settings.SERVER_SESSION_CLASS)
    if not _AccountDB:
        from evennia.accounts.models import AccountDB as _AccountDB
    if not _ServerConfig:
        from evennia.server.models import ServerConfig as _ServerConfig
    if not _ScriptDB:
        from evennia.scripts.models import ScriptDB as _ScriptDB
    # including once to avoid warnings in Python syntax checkers
    assert _ServerSession, "ServerSession class could not load"
    assert _AccountDB, "AccountDB class could not load"
    assert _ServerConfig, "ServerConfig class could not load"
    assert _ScriptDB, "ScriptDB class c ould not load"


# -----------------------------------------------------------
# SessionHandler base class
# ------------------------------------------------------------


class SessionHandler(dict):
    """
    This handler holds a stack of sessions.

    """

    def __getitem__(self, key):
        """
        Clean out None-sessions automatically.

        """
        if None in self:
            del self[None]
        return super().__getitem__(key)

    def get(self, key, default=None):
        """
        Clean out None-sessions automatically.

        """
        if None in self:
            del self[None]
        return super().get(key, default)

    def __setitem__(self, key, value):
        """
        Don't assign None sessions"

        """
        if key is not None:
            super().__setitem__(key, value)

    def __contains__(self, key):
        """
        None-keys are not accepted.

        """
        return False if key is None else super().__contains__(key)

    def get_sessions(self, include_unloggedin=False):
        """
        Returns the connected session objects.

        Args:
            include_unloggedin (bool, optional): Also list Sessions
                that have not yet authenticated.

        Returns:
            sessions (list): A list of `Session` objects.

        """
        if include_unloggedin:
            return list(self.values())
        else:
            return [session for session in self.values() if session.logged_in]

    def get_all_sync_data(self):
        """
        Create a dictionary of sessdata dicts representing all
        sessions in store.

        Returns:
            syncdata (dict): A dict of sync data.

        """
        return dict((sessid, sess.get_sync_data()) for sessid, sess in self.items())

    def clean_senddata(self, session, kwargs):
        """
        Clean up data for sending across the AMP wire. Also apply the
        FuncParser using callables from `settings.FUNCPARSER_OUTGOING_MESSAGES_MODULES`.

        Args:
            session (Session): The relevant session instance.
            kwargs (dict) Each keyword represents a send-instruction, with the keyword itself being
                the name of the instruction (like "text"). Suitable values for each keyword are:
                - arg                ->  [[arg], {}]
                - [args]             ->  [[args], {}]
                - {kwargs}           ->  [[], {kwargs}]
                - [args, {kwargs}]   ->  [[arg], {kwargs}]
                - [[args], {kwargs}] ->  [[args], {kwargs}]

        Returns:
            kwargs (dict): A cleaned dictionary of cmdname:[[args],{kwargs}] pairs,
            where the keys, args and kwargs have all been converted to
            send-safe entities (strings or numbers), and funcparser parsing has been
            applied.

        """
        global _FUNCPARSER
        if not _FUNCPARSER:
            from evennia.utils.funcparser import FuncParser

            _FUNCPARSER = FuncParser(
                settings.FUNCPARSER_OUTGOING_MESSAGES_MODULES, raise_errors=True
            )

        options = kwargs.pop("options", None) or {}
        raw = options.get("raw", False)
        strip_inlinefunc = options.get("strip_inlinefunc", False)

        def _utf8(data):
            if isinstance(data, bytes):
                try:
                    data = codecs_decode(data, session.protocol_flags["ENCODING"])
                except LookupError:
                    # wrong encoding set on the session. Set it to a safe one
                    session.protocol_flags["ENCODING"] = "utf-8"
                    data = codecs_decode(data, "utf-8")
                except UnicodeDecodeError:
                    # incorrect unicode sequence
                    session.sendLine(_ERR_BAD_UTF8)
                    data = ""

            return data

        def _validate(data):
            """
            Helper function to convert data to AMP-safe (picketable) values"

            """
            if isinstance(data, dict):
                newdict = {}
                for key, part in data.items():
                    newdict[key] = _validate(part)
                return newdict
            elif is_iter(data):
                return [_validate(part) for part in data]
            elif isinstance(data, (str, bytes)):
                data = _utf8(data)

                if (
                    _FUNCPARSER_PARSE_OUTGOING_MESSAGES_ENABLED
                    and not raw
                    and isinstance(self, ServerSessionHandler)
                ):
                    # only apply funcparser on the outgoing path (sessionhandler->)
                    # data = parse_inlinefunc(data, strip=strip_inlinefunc, session=session)
                    data = _FUNCPARSER.parse(data, strip=strip_inlinefunc, session=session)

                return str(data)
            elif (
                hasattr(data, "id")
                and hasattr(data, "db_date_created")
                and hasattr(data, "__dbclass__")
            ):
                # convert database-object to their string representation.
                return _validate(str(data))
            else:
                return data

        rkwargs = {}
        for key, data in kwargs.items():
            key = _validate(key)
            if not data:
                if key == "text":
                    # we don't allow sending text = None, this must mean
                    # that the text command is not to be used.
                    continue
                rkwargs[key] = [[], {}]
            elif isinstance(data, dict):
                rkwargs[key] = [[], _validate(data)]
            elif is_iter(data):
                data = tuple(data)
                if isinstance(data[-1], dict):
                    if len(data) == 2:
                        if is_iter(data[0]):
                            rkwargs[key] = [_validate(data[0]), _validate(data[1])]
                        else:
                            rkwargs[key] = [[_validate(data[0])], _validate(data[1])]
                    else:
                        rkwargs[key] = [_validate(data[:-1]), _validate(data[-1])]
                else:
                    rkwargs[key] = [_validate(data), {}]
            else:
                rkwargs[key] = [[_validate(data)], {}]
            rkwargs[key][1]["options"] = dict(options)
        # make sure that any "prompt" message will be processed last
        # by moving it to the end
        if "prompt" in rkwargs:
            prompt = rkwargs.pop("prompt")
            rkwargs["prompt"] = prompt
        return rkwargs


# ------------------------------------------------------------
# Server-SessionHandler class
# ------------------------------------------------------------


class ServerSessionHandler(SessionHandler):
    """
    This object holds the stack of sessions active in the game at any time.

    A session register with the handler in two steps, first by registering itself with the connect()
    method. This indicates an non-authenticated session. Whenever the session is authenticated the
    session together with the related account is sent to the login() method.

    """

    # AMP communication methods

    def __init__(self, *args, **kwargs):
        """
        Init the handler.

        """
        super().__init__(*args, **kwargs)
        evennia.server_data = {"servername": _SERVERNAME}
        # will be set on psync
        self.portal_start_time = 0.0

    def _run_cmd_login(self, session):
        """
        Launch the CMD_LOGINSTART command. This is wrapped
        for delays.

        """
        if not session.logged_in:
            self.data_in(session, text=[[CMD_LOGINSTART], {}])

    def portal_connect(self, portalsessiondata):
        """
        Called by Portal when a new session has connected.
        Creates a new, unlogged-in game session.

        Args:
            portalsessiondata (dict): a dictionary of all property:value
                keys defining the session and which is marked to be
                synced.

        """
        delayed_import()
        global _ServerSession, _AccountDB, _ScriptDB

        sess = _ServerSession()
        sess.sessionhandler = self
        sess.load_sync_data(portalsessiondata)
        sess.at_sync()
        # validate all scripts
        # _ScriptDB.objects.validate()
        self[sess.sessid] = sess

        if sess.logged_in and sess.uid:
            # Session is already logged in. This can happen in the
            # case of auto-authenticating protocols like SSH or
            # webclient's session sharing
            account = _AccountDB.objects.get_account_from_uid(sess.uid)
            if account:
                # this will set account.is_connected too
                self.login(sess, account, force=True)
                return
            else:
                sess.logged_in = False
                sess.uid = None

        # show the first login command, may delay slightly to allow
        # the handshakes to finish.
        delay(_DELAY_CMD_LOGINSTART, self._run_cmd_login, sess)

    def portal_session_sync(self, portalsessiondata):
        """
        Called by Portal when it wants to update a single session (e.g.
        because of all negotiation protocols have finally replied)

        Args:
            portalsessiondata (dict): a dictionary of all property:value
                keys defining the session and which is marked to be
                synced.

        """
        sessid = portalsessiondata.get("sessid")
        session = self.get(sessid)
        if session:
            # since some of the session properties may have had
            # a chance to change already before the portal gets here
            # the portal doesn't send all sessiondata but only
            # ones which should only be changed from portal (like
            # protocol_flags etc)
            session.load_sync_data(portalsessiondata)

    def portal_sessions_sync(self, portalsessionsdata):
        """
        Syncing all session ids of the portal with the ones of the
        server. This is instantiated by the portal when reconnecting.

        Args:
            portalsessionsdata (dict): A dictionary
              `{sessid: {property:value},...}` defining each session and
              the properties in it which should be synced.

        """
        delayed_import()
        global _ServerSession, _AccountDB, _ServerConfig, _ScriptDB

        for sess in list(self.values()):
            # we delete the old session to make sure to catch eventual
            # lingering references.
            del sess

        for sessid, sessdict in portalsessionsdata.items():
            sess = _ServerSession()
            sess.sessionhandler = self
            sess.load_sync_data(sessdict)
            if sess.uid:
                sess.account = _AccountDB.objects.get_account_from_uid(sess.uid)
            self[sessid] = sess
            sess.at_sync()

        mode = "reload"

        # tell the server hook we synced
        evennia.EVENNIA_SERVER_SERVICE.at_post_portal_sync(mode)
        # announce the reconnection
        if _BROADCAST_SERVER_RESTART_MESSAGES:
            self.announce_all(_(" ... Server restarted."))

    def portal_disconnect(self, session):
        """
        Called from Portal when Portal session closed from the portal
        side. There is no message to report in this case.

        Args:
            session (Session): The Session to disconnect

        """
        # disconnect us without calling Portal since
        # Portal already knows.
        self.disconnect(session, reason="", sync_portal=False)

    def portal_disconnect_all(self):
        """
        Called from Portal when Portal is closing down. All
        Sessions should die. The Portal should not be informed.

        """
        # set a watchdog to avoid self.disconnect from deleting
        # the session while we are looping over them
        self._disconnect_all = True
        for session in self.values():
            session.disconnect()
        del self._disconnect_all

    # server-side access methods

    def start_bot_session(self, protocol_path, configdict):
        """
        This method allows the server-side to force the Portal to
        create a new bot session.

        Args:
            protocol_path (str): The  full python path to the bot's
                class.
            configdict (dict): This dict will be used to configure
                the bot (this depends on the bot protocol).

        Examples:
            start_bot_session("evennia.server.portal.irc.IRCClient",
                              {"uid":1,  "botname":"evbot", "channel":"#evennia",
                               "network:"irc.freenode.net", "port": 6667})

        Notes:
            The new session will use the supplied account-bot uid to
            initiate an already logged-in connection. The Portal will
            treat this as a normal connection and henceforth so will
            the Server.

        """
        evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.SCONN, protocol_path=protocol_path, config=configdict
        )

    def portal_restart_server(self):
        """
        Called by server when reloading. We tell the portal to start a new server instance.

        """
        evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.SRELOAD
        )

    def portal_reset_server(self):
        """
        Called by server when reloading. We tell the portal to start a new server instance.

        """
        evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.SRESET
        )

    def portal_shutdown(self):
        """
        Called by server when it's time to shut down (the portal will shut us down and then shut
        itself down)

        """
        evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.PSHUTD
        )

    def login(self, session, account, force=False, testmode=False):
        """
        Log in the previously unloggedin session and the account we by now should know is connected
        to it. After this point we assume the session to be logged in one way or another.

        Args:
            session (Session): The Session to authenticate.
            account (Account): The Account identified as associated with this Session.
            force (bool): Login also if the session thinks it's already logged in
                (this can happen for auto-authenticating protocols)
            testmode (bool, optional): This is used by unittesting for
                faking login without any AMP being actually active.

        """
        if session.logged_in and not force:
            # don't log in a session that is already logged in.
            return

        account.is_connected = True

        # sets up and assigns all properties on the session
        session.at_login(account)

        # account init
        account.at_init()

        # Check if this is the first time the *account* logs in
        if account.db.FIRST_LOGIN:
            account.at_first_login()
            del account.db.FIRST_LOGIN

        account.at_pre_login()

        if _MULTISESSION_MODE == 0:
            # disconnect all previous sessions.
            self.disconnect_duplicate_sessions(session)

        nsess = len(self.sessions_from_account(account))
        string = "Logged in: {account} {address} ({nsessions} session(s) total)"
        string = string.format(account=account, address=session.address, nsessions=nsess)
        session.log(string)
        session.logged_in = True
        # sync the portal to the session
        if not testmode:
            evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
                session, operation=amp.SLOGIN, sessiondata={"logged_in": True, "uid": session.uid}
            )
        account.at_post_login(session=session)
        if nsess < 2:
            SIGNAL_ACCOUNT_POST_FIRST_LOGIN.send(sender=account, session=session)
        SIGNAL_ACCOUNT_POST_LOGIN.send(sender=account, session=session)

    def disconnect(self, session, reason="", sync_portal=True):
        """
        Called from server side to remove session and inform portal
        of this fact.

        Args:
            session (Session): The Session to disconnect.
            reason (str, optional): A motivation for the disconnect.
            sync_portal (bool, optional): Sync the disconnect to
                Portal side. This should be done unless this was
                called by self.portal_disconnect().

        """
        session = self.get(session.sessid)
        if not session:
            return

        if hasattr(session, "account") and session.account:
            # only log accounts logging off
            nsess = len(self.sessions_from_account(session.account)) - 1
            sreason = " ({})".format(reason) if reason else ""
            string = "Logged out: {account} {address} ({nsessions} sessions(s) remaining){reason}"
            string = string.format(
                reason=sreason, account=session.account, address=session.address, nsessions=nsess
            )
            session.log(string)

            if nsess == 0:
                SIGNAL_ACCOUNT_POST_LAST_LOGOUT.send(sender=session.account, session=session)

        session.at_disconnect(reason)
        SIGNAL_ACCOUNT_POST_LOGOUT.send(sender=session.account, session=session)
        sessid = session.sessid
        if sessid in self and not hasattr(self, "_disconnect_all"):
            del self[sessid]
        if sync_portal:
            # inform portal that session should be closed.
            evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
                session, operation=amp.SDISCONN, reason=reason
            )

    def all_sessions_portal_sync(self):
        """
        This is called by the server when it reboots. It syncs all session data
        to the portal. Returns a deferred!

        """
        sessdata = self.get_all_sync_data()
        return evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.SSYNC, sessiondata=sessdata
        )

    def session_portal_sync(self, session):
        """
        This is called by the server when it wants to sync a single session
        with the Portal for whatever reason. Returns a deferred!

        """
        sessdata = {session.sessid: session.get_sync_data()}
        return evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.SSYNC, sessiondata=sessdata, clean=False
        )

    def session_portal_partial_sync(self, session_data):
        """
        Call to make a partial update of the session, such as only a particular property.

        Args:
            session_data (dict): Store `{sessid: {property:value}, ...}` defining one or
                more sessions in detail.

        """
        return evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.SSYNC, sessiondata=session_data, clean=False
        )

    def disconnect_all_sessions(self, reason="You have been disconnected."):
        """
        Cleanly disconnect all of the connected sessions.

        Args:
            reason (str, optional): The reason for the disconnection.

        """

        for session in self:
            del session
        # tell portal to disconnect all sessions
        evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_AdminServer2Portal(
            DUMMYSESSION, operation=amp.SDISCONNALL, reason=reason
        )

    def disconnect_duplicate_sessions(
        self, curr_session, reason=_("Logged in from elsewhere. Disconnecting.")
    ):
        """
        Disconnects any existing sessions with the same user.

        args:
            curr_session (Session): Disconnect all Sessions matching this one.
            reason (str, optional): A motivation for disconnecting.

        """
        uid = curr_session.uid
        # we can't compare sessions directly since this will compare addresses and
        # mean connecting from the same host would not catch duplicates
        sid = id(curr_session)
        doublet_sessions = [
            sess for sess in self.values() if sess.logged_in and sess.uid == uid and id(sess) != sid
        ]

        for session in doublet_sessions:
            self.disconnect(session, reason)

    def validate_sessions(self):
        """
        Check all currently connected sessions (logged in and not) and
        see if any are dead or idle.

        """
        tcurr = time.time()
        reason = _("Idle timeout exceeded, disconnecting.")
        for session in (
            session
            for session in self.values()
            if session.logged_in
            and _IDLE_TIMEOUT > 0
            and (tcurr - session.cmd_last) > _IDLE_TIMEOUT
        ):
            self.disconnect(session, reason=reason)

    def account_count(self):
        """
        Get the number of connected accounts (not sessions since a
        account may have more than one session depending on settings).
        Only logged-in accounts are counted here.

        Returns:
            naccount (int): Number of connected accounts

        """
        return len(set(session.uid for session in self.values() if session.logged_in))

    def all_connected_accounts(self):
        """
        Get a unique list of connected and logged-in Accounts.

        Returns:
            accounts (list): All connected Accounts (which may be fewer than the
                amount of Sessions due to multi-playing).

        """
        return list(
            set(
                session.account
                for session in self.values()
                if session.logged_in and session.account
            )
        )

    def session_from_sessid(self, sessid):
        """
        Get session based on sessid, or None if not found

        Args:
            sessid (int or list): Session id(s).

        Return:
            sessions (Session or list): Session(s) found. This
                is a list if input was a list.

        """
        if is_iter(sessid):
            return [self.get(sid) for sid in sessid if sid in self]
        return self.get(sessid)

    def session_from_account(self, account, sessid):
        """
        Given an account and a session id, return the actual session
        object.

        Args:
            account (Account): The Account to get the Session from.
            sessid (int or list): Session id(s).

        Returns:
            sessions (Session or list): Session(s) found.

        """
        sessions = [
            self[sid]
            for sid in make_iter(sessid)
            if sid in self and self[sid].logged_in and account.uid == self[sid].uid
        ]
        return sessions[0] if len(sessions) == 1 else sessions

    def sessions_from_account(self, account):
        """
        Given an account, return all matching sessions.

        Args:
            account (Account): Account to get sessions from.

        Returns:
            sessions (list): All Sessions associated with this account.

        """
        uid = account.uid
        return [session for session in self.values() if session.logged_in and session.uid == uid]

    def sessions_from_puppet(self, puppet):
        """
        Given a puppeted object, return all controlling sessions.

        Args:
            puppet (Object): Object puppeted

        Returns.
            sessions (Session or list): Can be more than one of Object is controlled by more than
                one Session (MULTISESSION_MODE > 1).

        """
        sessions = puppet.sessid.get()
        return sessions[0] if len(sessions) == 1 else sessions

    sessions_from_character = sessions_from_puppet

    def sessions_from_csessid(self, csessid):
        """
        Given a client identification hash (for session types that offer them)
        return all sessions with a matching hash.

        Args
            csessid (str): The session hash.

        Returns:
            sessions (list): The sessions with matching .csessid, if any.

        """
        if not csessid:
            return []
        return [
            session for session in self.values() if session.csessid and session.csessid == csessid
        ]

    def announce_all(self, message):
        """
        Send message to all connected sessions

        Args:
            message (str): Message to send.

        """
        for session in self.values():
            self.data_out(session, text=message)

    def data_out(self, session, **kwargs):
        """
        Sending data Server -> Portal

        Args:
            session (Session): Session to relay to.
            text (str, optional): text data to return

        Notes:
            The outdata will be scrubbed for sending across
            the wire here.
        """
        # clean output for sending
        kwargs = self.clean_senddata(session, kwargs)

        # send across AMP
        evennia.EVENNIA_SERVER_SERVICE.amp_protocol.send_MsgServer2Portal(session, **kwargs)

    def get_inputfuncs(self):
        """
        Get all registered inputfuncs (access function)

        Returns:
            inputfuncs (dict): A dict of {key:inputfunc,...}
        """
        return _INPUT_FUNCS

    def data_in(self, session, **kwargs):
        """
        We let the data take a "detour" to session.data_in
        so the user can override and see it all in one place.
        That method is responsible to in turn always call
        this class' `sessionhandler.call_inputfunc` with the
        (possibly processed) data.

        """
        if session:
            session.data_in(**kwargs)

    def call_inputfuncs(self, session, **kwargs):
        """
        Split incoming data into its inputfunc counterparts. This should be
        called by the `serversession.data_in` as
        `sessionhandler.call_inputfunc(self, **kwargs)`.

        We also intercept OOB communication here.

        Args:
            sessions (Session): Session.

        Keyword Args:
            any (tuple): Incoming data from protocol, each
                on the form `commandname=((args), {kwargs})`.

        """

        # distribute incoming data to the correct receiving methods.
        if session:
            input_debug = session.protocol_flags.get("INPUTDEBUG", False)
            for cmdname, (cmdargs, cmdkwargs) in kwargs.items():
                cname = cmdname.strip().lower()
                try:
                    cmdkwargs.pop("options", None)
                    if cname in _INPUT_FUNCS:
                        _INPUT_FUNCS[cname](session, *cmdargs, **cmdkwargs)
                    else:
                        _INPUT_FUNCS["default"](session, cname, *cmdargs, **cmdkwargs)
                except Exception as err:
                    if input_debug:
                        session.msg(err)
                    log_trace()


# These are filled in during server boot.
SESSION_HANDLER = None
SESSIONS = None  # legacy



# -*- coding: utf-8 -*-


import django.core.validators
from django.db import migrations, models

import evennia.accounts.manager


class Migration(migrations.Migration):

    dependencies = [("accounts", "0003_auto_20150209_2234")]

    operations = [
        migrations.DeleteModel(name="DefaultGuest"),
        migrations.DeleteModel(name="DefaultAccount"),
        migrations.AlterModelManagers(
            name="accountdb", managers=[("objects", evennia.accounts.manager.AccountDBManager())]
        ),
        migrations.AlterField(
            model_name="accountdb",
            name="email",
            field=models.EmailField(max_length=254, verbose_name="email address", blank=True),
        ),
        migrations.AlterField(
            model_name="accountdb",
            name="groups",
            field=models.ManyToManyField(
                related_query_name="user",
                related_name="user_set",
                to="auth.Group",
                blank=True,
                help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                verbose_name="groups",
            ),
        ),
        migrations.AlterField(
            model_name="accountdb",
            name="last_login",
            field=models.DateTimeField(null=True, verbose_name="last login", blank=True),
        ),
        migrations.AlterField(
            model_name="accountdb",
            name="username",
            field=models.CharField(
                error_messages={"unique": "A user with that username already exists."},
                max_length=30,
                validators=[
                    django.core.validators.RegexValidator(
                        "^[\\w.@+-]+$",
                        "Enter a valid username. This value may contain only letters, numbers and @/./+/-/_ characters.",
                        "invalid",
                    )
                ],
                help_text="Required. 30 characters or fewer. Letters, digits and @/./+/-/_ only.",
                unique=True,
                verbose_name="username",
            ),
        ),
    ]



"""
Test the main server component

"""
from unittest import TestCase

from django.test import override_settings
from mock import DEFAULT, MagicMock, call, patch

import evennia


@patch("evennia.server.service.LoopingCall", new=MagicMock())
class TestServer(TestCase):
    """
    Test server module.

    """

    def setUp(self):
        # Running this first line ensures that the EVENNIA_SERVICE is instantiated.
        from evennia.server import server

        self.server = evennia.EVENNIA_SERVER_SERVICE

    @override_settings(IDMAPPER_CACHE_MAXSIZE=1000)
    def test__server_maintenance_reset(self):
        with patch.object(self.server, "_flush_cache", new=MagicMock()) as mockflush, patch.object(
            evennia, "ServerConfig", new=MagicMock()
        ) as mockconf, patch.multiple(
            "evennia.server.service",
            LoopingCall=DEFAULT,
            connection=DEFAULT,
        ) as mocks:
            self.server.maintenance_count = 0

            mocks["connection"].close = MagicMock()
            mockconf.objects.conf = MagicMock(return_value=456)

            # flush cache
            self.server.server_maintenance()
            mockconf.objects.conf.assert_called_with("runtime", 456)

    @override_settings(IDMAPPER_CACHE_MAXSIZE=1000)
    def test__server_maintenance_flush(self):
        with patch.multiple(
            "evennia.server.service",
            LoopingCall=DEFAULT,
            connection=DEFAULT,
        ) as mocks, patch.object(
            evennia, "ServerConfig", new=MagicMock()
        ) as mockconf, patch.object(
            self.server, "_flush_cache", new=MagicMock()
        ) as mockflush:
            mocks["connection"].close = MagicMock()
            mockconf.objects.conf = MagicMock(return_value=100)
            self.server.maintenance_count = 5 - 1
            # flush cache
            self.server.server_maintenance()
            self.server._flush_cache.assert_called_with(1000)

    @override_settings(IDMAPPER_CACHE_MAXSIZE=1000)
    def test__server_maintenance_close_connection(self):
        with patch.multiple(
            "evennia.server.service",
            LoopingCall=DEFAULT,
            connection=DEFAULT,
        ) as mocks, patch.object(evennia, "ServerConfig", new=MagicMock()) as mockconf:
            self.server._flush_cache = MagicMock()
            self.server.maintenance_count = (60 * 7) - 1
            self.server._last_server_time_snapshot = 0
            mocks["connection"].close = MagicMock()
            mockconf.objects.conf = MagicMock(return_value=100)
            self.server.server_maintenance()
            mocks["connection"].close.assert_called()

    @override_settings(IDLE_TIMEOUT=10)
    def test__server_maintenance_idle_time(self):
        with patch.multiple(
            "evennia.server.service",
            LoopingCall=DEFAULT,
            connection=DEFAULT,
            time=DEFAULT,
        ) as mocks, patch.object(
            evennia, "ServerConfig", new=MagicMock()
        ) as mockconf, patch.object(
            evennia, "SESSION_HANDLER", new=MagicMock()
        ) as mocksess:
            self.server.maintenance_count = (3600 * 7) - 1
            self.server._last_server_time_snapshot = 0
            sess1 = MagicMock()
            sess2 = MagicMock()
            sess3 = MagicMock()
            sess4 = MagicMock()
            sess1.cmd_last = 100  # should time out
            sess2.cmd_last = 999  # should not time out
            sess3.cmd_last = 100  # should not time (due to account)
            sess4.cmd_last = 100  # should time out (due to access)
            sess1.account = None
            sess2.account = None
            sess3.account = MagicMock()
            sess3.account = MagicMock()
            sess4.account.access = MagicMock(return_value=False)

            mocks["time"].time = MagicMock(return_value=1000)

            mockconf.objects.conf = MagicMock(return_value=100)
            mocksess.values = MagicMock(return_value=[sess1, sess2, sess3, sess4])
            mocksess.disconnect = MagicMock()

            self.server.server_maintenance()
            reason = "idle timeout exceeded"
            calls = [call(sess1, reason=reason), call(sess4, reason=reason)]
            mocksess.disconnect.assert_has_calls(calls, any_order=True)

    def test_update_defaults(self):
        with patch.object(evennia, "ObjectDB", new=MagicMock()) as mockobj, patch.object(
            evennia, "AccountDB", new=MagicMock()
        ) as mockacc, patch.object(evennia, "ScriptDB", new=MagicMock()) as mockscr, patch.object(
            evennia, "ChannelDB", new=MagicMock()
        ) as mockchan, patch.object(
            evennia, "ServerConfig", new=MagicMock()
        ) as mockconf:
            for m in (mockscr, mockobj, mockacc, mockchan):
                m.objects.filter = MagicMock()

            # fake mismatches
            settings_names = (
                "CMDSET_CHARACTER",
                "CMDSET_ACCOUNT",
                "BASE_ACCOUNT_TYPECLASS",
                "BASE_OBJECT_TYPECLASS",
                "BASE_CHARACTER_TYPECLASS",
                "BASE_ROOM_TYPECLASS",
                "BASE_EXIT_TYPECLASS",
                "BASE_SCRIPT_TYPECLASS",
                "BASE_CHANNEL_TYPECLASS",
            )
            fakes = {name: "Dummy.path" for name in settings_names}

            def _mock_conf(key, *args):
                return fakes[key]

            mockconf.objects.conf = _mock_conf

            self.server.update_defaults()

            for m in (mockscr, mockobj, mockacc, mockchan):
                m.objects.filter.assert_called()

    @override_settings(TEST_ENVIRONMENT=True)
    def test_initial_setup(self):
        from evennia.utils.create import create_account

        acct = create_account("TestSuperuser", "test@test.com", "testpassword", is_superuser=True)

        with patch.multiple(
            "evennia.server.initial_setup", reset_server=DEFAULT, AccountDB=DEFAULT
        ) as mocks:
            mocks["AccountDB"].objects.get = MagicMock(return_value=acct)
            self.server.run_initial_setup()
        acct.delete()

    @override_settings(TEST_ENVIRONMENT=True)
    def test_initial_setup_retry(self):
        from evennia.utils.create import create_account

        acct = create_account("TestSuperuser2", "test@test.com", "testpassword", is_superuser=True)

        with patch.multiple(
            "evennia.server.initial_setup",
            ServerConfig=DEFAULT,
            reset_server=DEFAULT,
            AccountDB=DEFAULT,
        ) as mocks:
            mocks["AccountDB"].objects.get = MagicMock(return_value=acct)
            # a last_initial_setup_step > 0
            mocks["ServerConfig"].objects.conf = MagicMock(return_value=4)
            self.server.run_initial_setup()
        acct.delete()

    def test_get_info_dict(self):
        with patch.object(self.server, "get_info_dict", return_value={"test": "foo"}) as mocks:
            self.assertEqual(self.server.get_info_dict(), {"test": "foo"})


class TestInitHooks(TestCase):
    def setUp(self):
        from evennia.server import server
        from evennia.utils import create

        self.server = evennia.EVENNIA_SERVER_SERVICE

        self.obj1 = create.object(key="HookTestObj1")
        self.obj2 = create.object(key="HookTestObj2")
        self.acct1 = create.account("HookAcct1", "hooktest1@test.com", "testpasswd")
        self.acct2 = create.account("HookAcct2", "hooktest2@test.com", "testpasswd")
        self.chan1 = create.channel("Channel1")
        self.chan2 = create.channel("Channel2")
        self.script1 = create.script(key="script1")
        self.script2 = create.script(key="script2")

        self.objects = [
            self.obj1,
            self.obj2,
            self.acct1,
            self.acct2,
            self.chan1,
            self.chan2,
            self.script1,
            self.script2,
        ]

        for obj in self.objects:
            obj.at_init = MagicMock()

    def tearDown(self):
        for obj in self.objects:
            obj.delete()

    @override_settings(TEST_ENVIRONMENT=True)
    def test_run_init_hooks(self):
        with patch.object(
            self.server, "at_server_reload_start", new=MagicMock()
        ) as reload, patch.object(self.server, "at_server_cold_start", new=MagicMock()) as cold:
            self.server.run_init_hooks("reload")
            self.server.run_init_hooks("reset")
            self.server.run_init_hooks("shutdown")

            for obj in self.objects:
                obj.at_init.assert_called()

            for hook in (reload, cold):
                hook.assert_called()



"""
Testing of TestDice.

"""

from mock import patch

from evennia.commands.default.tests import BaseEvenniaCommandTest

from . import dice


@patch("evennia.contrib.rpg.dice.dice.randint", return_value=5)
class TestDice(BaseEvenniaCommandTest):
    def test_roll_dice(self, mocked_randint):
        self.assertEqual(dice.roll(6, 6, modifier=("+", 4)), mocked_randint() * 6 + 4)
        self.assertEqual(dice.roll(6, 6, conditional=("<", 35)), True)
        self.assertEqual(dice.roll(6, 6, conditional=(">", 33)), False)

    def test_cmddice(self, mocked_randint):
        self.call(
            dice.CmdDice(), "3d6 + 4", "You roll 3d6 + 4.| Roll(s): 5, 5 and 5. Total result is 19."
        )
        self.call(dice.CmdDice(), "100000d1000", "The maximum roll allowed is 10000d10000.")
        self.call(dice.CmdDice(), "/secret 3d6 + 4", "You roll 3d6 + 4 (secret, not echoed).")

    def test_string_form(self, mocked_randint):
        self.assertEqual(dice.roll("6d6 + 4"), mocked_randint() * 6 + 4)
        self.assertEqual(dice.roll("6d6 < 35"), True)
        self.assertEqual(dice.roll("6d6 > 35"), False)
        self.assertEqual(dice.roll("2d10 + 5 >= 14"), True)

    def test_maxvals(self, mocked_randint):
        with self.assertRaises(TypeError):
            dice.roll(11, 1001, max_dicenum=10, max_dicetype=1000)
        with self.assertRaises(TypeError):
            dice.roll(10, 1001, max_dicenum=10, max_dicetype=1000)



"""
Unit tests for the EvMenu system

This sets up a testing parent for testing EvMenu trees. It is configured by subclassing the
`TestEvMenu` class from this module and setting the class variables to point to the menu that should
be tested and how it should be called.

Without adding any further test methods, the tester will process all nodes of the menu, depth first,
by stepping through all options for every node. Optionally, it can check that all nodes are visited.
It will create a hierarchical list of node names that describes the tree structure. This can then be
compared against a template to make sure the menu structure is sound. Easiest way to use this is to
run the test once to see how the structure looks.

The system also allows for testing the returns of each node as part of the parsing.

To help debug the menu, turn on `debug_output`, which will print the traversal process in detail.

"""

import copy

from anything import Anything
from django.test import TestCase
from mock import MagicMock

from evennia.utils import ansi, evmenu
from evennia.utils.test_resources import BaseEvenniaTest


class TestEvMenu(TestCase):
    "Run the EvMenu testing."
    menutree = {}  # can also be the path to the menu tree
    startnode = "start"
    cmdset_mergetype = "Replace"
    cmdset_priority = 1
    auto_quit = True
    auto_look = True
    auto_help = True
    cmd_on_exit = "look"
    persistent = False
    startnode_input = ""
    kwargs = {}

    # if all nodes must be visited for the test to pass. This is not on
    # by default since there may be exec-nodes that are made to not be
    # visited.
    expect_all_nodes = False

    # this is compared against the full tree structure generated
    expected_tree = []
    # this allows for verifying that a given node returns a given text. The
    # text is compared with .startswith, so the entire text need not be matched.
    expected_node_texts = {}
    # just check the number of options from each node
    expected_node_options_count = {}
    # check the actual options
    expected_node_options = {}

    # set this to print the traversal as it happens (debugging)
    debug_output = False

    def _debug_output(self, indent, msg):
        if self.debug_output:
            print(" " * indent + ansi.strip_ansi(msg))

    def _test_menutree(self, menu):
        """
        This is a automatic tester of the menu tree by recursively progressing through the
        structure.
        """

        def _depth_first(menu, tree, visited, indent):
            # we are in a given node here
            nodename = menu.nodename
            options = menu.test_options
            if isinstance(options, dict):
                options = (options,)

            # run validation tests for this node
            compare_text = self.expected_node_texts.get(nodename, None)
            if compare_text is not None:
                compare_text = ansi.strip_ansi(compare_text.strip())
                node_text = menu.test_nodetext
                self.assertIsNotNone(
                    bool(node_text),
                    "node: {}: node-text is None, which was not expected.".format(nodename),
                )
                if isinstance(node_text, tuple):
                    node_text, helptext = node_text
                node_text = ansi.strip_ansi(node_text.strip())
                self.assertTrue(
                    node_text.startswith(compare_text),
                    "\nnode \"{}':\nOutput:\n{}\n\nExpected (startswith):\n{}".format(
                        nodename, node_text, compare_text
                    ),
                )
            compare_options_count = self.expected_node_options_count.get(nodename, None)
            if compare_options_count is not None:
                self.assertEqual(
                    len(options),
                    compare_options_count,
                    "Not the right number of options returned from node {}.".format(nodename),
                )
            compare_options = self.expected_node_options.get(nodename, None)
            if compare_options:
                self.assertEqual(
                    options,
                    compare_options,
                    "Options returned from node {} does not match.".format(nodename),
                )

            self._debug_output(indent, "*{}".format(nodename))
            subtree = []

            if not options:
                # an end node
                if nodename not in visited:
                    visited.append(nodename)
                subtree = nodename
            else:
                for inum, optdict in enumerate(options):
                    key, desc, execute, goto = (
                        optdict.get("key", ""),
                        optdict.get("desc", None),
                        optdict.get("exec", None),
                        optdict.get("goto", None),
                    )

                    # prepare the key to pass to the menu
                    if isinstance(key, (tuple, list)) and len(key) > 1:
                        key = key[0]
                    if key == "_default":
                        key = "test raw input"
                    if not key:
                        key = str(inum + 1)

                    backup_menu = copy.copy(menu)

                    # step the menu
                    menu.parse_input(key)

                    # from here on we are likely in a different node
                    nodename = menu.nodename

                    if menu.close_menu.called:
                        # this was an end node
                        self._debug_output(indent, "    .. menu exited! Back to previous node.")
                        menu = backup_menu
                        menu.close_menu = MagicMock()
                        visited.append(nodename)
                        subtree.append(nodename)
                    elif nodename not in visited:
                        visited.append(nodename)
                        subtree.append(nodename)
                        _depth_first(menu, subtree, visited, indent + 2)
                        # self._debug_output(indent, "    -> arrived at {}".format(nodename))
                    else:
                        subtree.append(nodename)
                        # self._debug_output( indent, "    -> arrived at {} (circular call)".format(nodename))
                    self._debug_output(indent, "-- {} ({}) -> {}".format(key, desc, goto))

            if subtree:
                tree.append(subtree)

        # the start node has already fired at this point
        visited_nodes = [menu.nodename]
        traversal_tree = [menu.nodename]
        _depth_first(menu, traversal_tree, visited_nodes, 1)

        if self.expect_all_nodes:
            self.assertGreaterEqual(len(menu._menutree), len(visited_nodes))
        self.assertEqual(traversal_tree, self.expected_tree)

    def setUp(self):
        self.menu = None
        if self.menutree:
            self.caller = MagicMock()
            self.caller.key = "Test"
            self.caller2 = MagicMock()
            self.caller2.key = "Test"
            self.caller.msg = MagicMock()
            self.caller2.msg = MagicMock()
            self.session = MagicMock()
            self.session.protocol_flags = {}
            self.session2 = MagicMock()
            self.session2.protocol_flags = {}
            self.caller.session = self.session
            self.caller2.session = self.session2

            self.menu = evmenu.EvMenu(
                self.caller,
                self.menutree,
                startnode=self.startnode,
                cmdset_mergetype=self.cmdset_mergetype,
                cmdset_priority=self.cmdset_priority,
                auto_quit=self.auto_quit,
                auto_look=self.auto_look,
                auto_help=self.auto_help,
                cmd_on_exit=self.cmd_on_exit,
                persistent=False,
                startnode_input=self.startnode_input,
                session=self.session,
                **self.kwargs,
            )
            # persistent version
            self.pmenu = evmenu.EvMenu(
                self.caller2,
                self.menutree,
                startnode=self.startnode,
                cmdset_mergetype=self.cmdset_mergetype,
                cmdset_priority=self.cmdset_priority,
                auto_quit=self.auto_quit,
                auto_look=self.auto_look,
                auto_help=self.auto_help,
                cmd_on_exit=self.cmd_on_exit,
                persistent=True,
                startnode_input=self.startnode_input,
                session=self.session2,
                **self.kwargs,
            )

            self.menu.close_menu = MagicMock()
            self.pmenu.close_menu = MagicMock()

    def test_menu_structure(self):
        if self.menu:
            self._test_menutree(self.menu)
            self._test_menutree(self.pmenu)


class TestEvMenuExample(TestEvMenu):
    menutree = "evennia.utils.tests.data.evmenu_example"
    startnode = "test_start_node"
    kwargs = {"testval": "val", "testval2": "val2"}
    debug_output = False

    expected_node_texts = {"test_view_node": "Your name is"}

    expected_tree = [
        "test_start_node",
        [
            "test_set_node",
            ["test_start_node"],
            "test_look_node",
            ["test_start_node"],
            "test_view_node",
            ["test_start_node"],
            "test_dynamic_node",
            [
                "test_dynamic_node",
                "test_dynamic_node",
                "test_dynamic_node",
                "test_dynamic_node",
                "test_start_node",
            ],
            "test_end_node",
            "test_displayinput_node",
            ["test_start_node"],
        ],
    ]

    def test_kwargsave(self):
        self.assertTrue(hasattr(self.menu, "testval"))
        self.assertTrue(hasattr(self.menu, "testval2"))


def _callnode1(caller, raw_string, **kwargs):
    return "node1"


def _callnode2(caller, raw_string, **kwargs):
    return "node2"


class TestMenuTemplateParse(BaseEvenniaTest):
    """Test menu templating helpers"""

    def setUp(self):
        super().setUp()

        self.menu_template = """
        ## node start

        Neque ea alias perferendis molestiae eligendi. Debitis exercitationem
        exercitationem quas blanditiis quisquam officia ut. Fugit aut fugit enim quia
        non. Earum et excepturi animi ex esse accusantium et. Id adipisci eos enim
        ratione.

        ## options

        1: first option -> node1
        2: second option -> node2
        next: node1

        ## node node1

        Node 1

        ## options

        fwd: node2
        call1: callnode1()
        call2: callnode2(foo=bar, bar=22, goo="another test")
        >: start

        ## node node2

        Text of node 2

        ## options

        > foo*: node1
        > [0-9]+?: node2
        > back: start

        """
        self.goto_callables = {"callnode1": _callnode1, "callnode2": _callnode2}

    def test_parse_menu_template(self):
        """EvMenu template testing"""

        menutree = evmenu.parse_menu_template(self.char1, self.menu_template, self.goto_callables)
        self.assertEqual(menutree, {"start": Anything, "node1": Anything, "node2": Anything})

    def test_template2menu(self):
        evmenu.template2menu(self.char1, self.menu_template, self.goto_callables)

    def test_parse_menu_fail(self):
        template = """
        ## NODE

        Text

        ## OPTIONS

        next: callnode2(invalid)
        """
        with self.assertRaises(RuntimeError):
            evmenu.parse_menu_template(self.char1, template, self.goto_callables)



"""
Unit tests for the utilities of the evennia.utils.gametime module.
"""

import time
import unittest
from datetime import datetime
from unittest.mock import Mock

from django.conf import settings
from django.test import TestCase

from evennia.utils import gametime


class TestGametime(TestCase):
    def setUp(self) -> None:
        self.time = time.time
        self._SERVER_EPOCH = gametime._SERVER_EPOCH
        time.time = Mock(return_value=datetime(2019, 4, 18, 13, 49, 38).timestamp())
        gametime._SERVER_EPOCH = None
        gametime.SERVER_RUNTIME = 600.0
        gametime.SERVER_START_TIME = time.time() - 300
        gametime.SERVER_RUNTIME_LAST_UPDATED = time.time() - 30
        gametime.TIMEFACTOR = 5.0
        self.timescripts = []

    def tearDown(self) -> None:
        time.time = self.time
        gametime._SERVER_EPOCH = self._SERVER_EPOCH
        gametime.SERVER_RUNTIME_LAST_UPDATED = 0.0
        gametime.SERVER_RUNTIME = 0.0
        gametime.SERVER_START_TIME = 0.0
        gametime.TIMEFACTOR = settings.TIME_FACTOR
        for script in self.timescripts:
            script.stop()

    def test_runtime(self):
        self.assertAlmostEqual(gametime.runtime(), 630.0)

    def test_server_epoch(self):
        self.assertAlmostEqual(gametime.server_epoch(), time.time() - 630.0)

    def test_uptime(self):
        self.assertAlmostEqual(gametime.uptime(), 300.0)

    def test_game_epoch_no_setting(self):
        self.assertAlmostEqual(gametime.game_epoch(), gametime.server_epoch())

    def test_game_epoch_setting(self):
        with self.settings(TIME_GAME_EPOCH=0):
            self.assertEqual(gametime.game_epoch(), 0)

    def test_gametime_simple(self):
        self.assertAlmostEqual(gametime.gametime(), 630.0 * 5)

    def test_gametime_absolute(self):
        self.assertAlmostEqual(
            gametime.gametime(absolute=True), datetime(2019, 4, 18, 14, 31, 38).timestamp()
        )

    def test_gametime_downtimes(self):
        gametime.IGNORE_DOWNTIMES = True
        self.assertAlmostEqual(gametime.gametime(), 630 * 5.0)
        gametime.IGNORE_DOWNTIMES = False

    def test_real_seconds_until(self):
        # using the gametime value above, we are working from the following
        # datetime: datetime.datetime(2019, 4, 18, 14, 31, 38, 245449)
        self.assertAlmostEqual(gametime.real_seconds_until(sec=48), 2)
        self.assertAlmostEqual(gametime.real_seconds_until(min=32), 12)
        self.assertAlmostEqual(gametime.real_seconds_until(hour=15), 720)
        self.assertAlmostEqual(gametime.real_seconds_until(day=19), 17280)
        self.assertAlmostEqual(gametime.real_seconds_until(month=5), 518400)
        self.assertAlmostEqual(gametime.real_seconds_until(year=2020), 6324480)

    def test_real_seconds_until_behind(self):
        self.assertAlmostEqual(gametime.real_seconds_until(sec=28), 10)
        self.assertAlmostEqual(gametime.real_seconds_until(min=30), 708)
        self.assertAlmostEqual(gametime.real_seconds_until(hour=13), 16560)
        self.assertAlmostEqual(gametime.real_seconds_until(day=17), 501120)
        self.assertAlmostEqual(gametime.real_seconds_until(month=1), 4752000)

    def test_real_seconds_until_leap_year(self):
        self.assertAlmostEqual(gametime.real_seconds_until(month=3), 5788800)

    def test_schedule(self):
        callback = Mock()
        script = gametime.schedule(callback, day=19)
        self.timescripts.append(script)
        self.assertIsInstance(script, gametime.TimeScript)
        self.assertAlmostEqual(script.interval, 17280)
        self.assertEqual(script.repeats, 1)

    def test_repeat_schedule(self):
        callback = Mock()
        script = gametime.schedule(callback, repeat=True, min=32)
        self.timescripts.append(script)
        self.assertIsInstance(script, gametime.TimeScript)
        self.assertAlmostEqual(script.interval, 12)
        self.assertEqual(script.repeats, 0)



"""
Module containing the commands of the in-game Python system.
"""

from datetime import datetime

from django.conf import settings

from evennia.contrib.base_systems.ingame_python.utils import get_event_handler
from evennia.utils.ansi import raw
from evennia.utils.eveditor import EvEditor
from evennia.utils.evtable import EvTable
from evennia.utils.utils import class_from_module, time_format

COMMAND_DEFAULT_CLASS = class_from_module(settings.COMMAND_DEFAULT_CLASS)

# Permissions
WITH_VALIDATION = getattr(settings, "callbackS_WITH_VALIDATION", None)
WITHOUT_VALIDATION = getattr(settings, "callbackS_WITHOUT_VALIDATION", "developer")
VALIDATING = getattr(settings, "callbackS_VALIDATING", "developer")

# Split help text
BASIC_HELP = "Add, edit or delete callbacks."

BASIC_USAGES = [
    "@call <object name> [= <callback name>]",
    "@call/add <object name> = <callback name> [parameters]",
    "@call/edit <object name> = <callback name> [callback number]",
    "@call/del <object name> = <callback name> [callback number]",
    "@call/tasks [object name [= <callback name>]]",
]

BASIC_SWITCHES = [
    "add    - add and edit a new callback",
    "edit   - edit an existing callback",
    "del    - delete an existing callback",
    "tasks  - show the list of differed tasks",
]

VALIDATOR_USAGES = ["@call/accept [object name = <callback name> [callback number]]"]

VALIDATOR_SWITCHES = ["accept - show callbacks to be validated or accept one"]

BASIC_TEXT = """
This command is used to manipulate callbacks.  A callback can be linked to
an object, to fire at a specific moment.  You can use the command without
switches to see what callbacks are active on an object:
  @call self
You can also specify a callback name if you want the list of callbacks
associated with this object of this name:
  @call north = can_traverse
You can also add a number after the callback name to see details on one callback:
  @call here = say 2
You can also add, edit or remove callbacks using the add, edit or del switches.
Additionally, you can see the list of differed tasks created by callbacks
(chained events to be called) using the /tasks switch.
"""

VALIDATOR_TEXT = """
You can also use this command to validate callbacks.  Depending on your game
setting, some users might be allowed to add new callbacks, but these callbacks
will not be fired until you accept them.  To see the callbacks needing
validation, enter the /accept switch without argument:
  @call/accept
A table will show you the callbacks that are not validated yet, who created
them and when.  You can then accept a specific callback:
  @call here = enter 1
Use the /del switch to remove callbacks that should not be connected.
"""


class CmdCallback(COMMAND_DEFAULT_CLASS):

    """
    Command to edit callbacks.
    """

    key = "@call"
    aliases = ["@callback", "@callbacks", "@calls"]
    locks = "cmd:perm({})".format(VALIDATING)
    if WITH_VALIDATION:
        locks += " or perm({})".format(WITH_VALIDATION)
    help_category = "Building"

    def get_help(self, caller, cmdset):
        """
        Return the help message for this command and this caller.

        The help text of this specific command will vary depending
        on user permission.

        Args:
            caller (Object or Account): the caller asking for help on the command.
            cmdset (CmdSet): the command set (if you need additional commands).

        Returns:
            docstring (str): the help text to provide the caller for this command.

        """
        lock = "perm({}) or perm(callbacks_validating)".format(VALIDATING)
        validator = caller.locks.check_lockstring(caller, lock)
        text = "\n" + BASIC_HELP + "\n\nUsages:\n  "

        # Usages
        text += "\n  ".join(BASIC_USAGES)
        if validator:
            text += "\n  " + "\n  ".join(VALIDATOR_USAGES)

        # Switches
        text += "\n\nSwitches:\n  "
        text += "\n  ".join(BASIC_SWITCHES)
        if validator:
            text += "\n  " + "\n  ".join(VALIDATOR_SWITCHES)

        # Text
        text += "\n" + BASIC_TEXT
        if validator:
            text += "\n" + VALIDATOR_TEXT

        return text

    def func(self):
        """Command body."""
        caller = self.caller
        lock = "perm({}) or perm(events_validating)".format(VALIDATING)
        validator = caller.locks.check_lockstring(caller, lock)
        lock = "perm({}) or perm(events_without_validation)".format(WITHOUT_VALIDATION)
        autovalid = caller.locks.check_lockstring(caller, lock)

        # First and foremost, get the callback handler and set other variables
        self.handler = get_event_handler()
        self.obj = None
        rhs = self.rhs or ""
        self.callback_name, sep, self.parameters = rhs.partition(" ")
        self.callback_name = self.callback_name.lower()
        self.is_validator = validator
        self.autovalid = autovalid
        if self.handler is None:
            caller.msg("The event handler is not running, can't " "access the event system.")
            return

        # Before the equal sign, there is an object name or nothing
        if self.lhs:
            self.obj = caller.search(self.lhs)
            if not self.obj:
                return

        # Switches are mutually exclusive
        switch = self.switches and self.switches[0] or ""
        if switch in ("", "add", "edit", "del") and self.obj is None:
            caller.msg("Specify an object's name or #ID.")
            return

        if switch == "":
            self.list_callbacks()
        elif switch == "add":
            self.add_callback()
        elif switch == "edit":
            self.edit_callback()
        elif switch == "del":
            self.del_callback()
        elif switch == "accept" and validator:
            self.accept_callback()
        elif switch in ["tasks", "task"]:
            self.list_tasks()
        else:
            caller.msg("Mutually exclusive or invalid switches were " "used, cannot proceed.")

    def list_callbacks(self):
        """Display the list of callbacks connected to the object."""
        obj = self.obj
        callback_name = self.callback_name
        parameters = self.parameters
        callbacks = self.handler.get_callbacks(obj)
        types = self.handler.get_events(obj)

        if callback_name:
            # Check that the callback name can be found in this object
            created = callbacks.get(callback_name)
            if created is None:
                self.msg("No callback {} has been set on {}.".format(callback_name, obj))
                return

            if parameters:
                # Check that the parameter points to an existing callback
                try:
                    number = int(parameters) - 1
                    assert number >= 0
                    callback = callbacks[callback_name][number]
                except (ValueError, AssertionError, IndexError):
                    self.msg(
                        "The callback {} {} cannot be found in {}.".format(
                            callback_name, parameters, obj
                        )
                    )
                    return

                # Display the callback's details
                author = callback.get("author")
                author = author.key if author else "|gUnknown|n"
                updated_by = callback.get("updated_by")
                updated_by = updated_by.key if updated_by else "|gUnknown|n"
                created_on = callback.get("created_on")
                created_on = (
                    created_on.strftime("%Y-%m-%d %H:%M:%S") if created_on else "|gUnknown|n"
                )
                updated_on = callback.get("updated_on")
                updated_on = (
                    updated_on.strftime("%Y-%m-%d %H:%M:%S") if updated_on else "|gUnknown|n"
                )
                msg = "Callback {} {} of {}:".format(callback_name, parameters, obj)
                msg += "\nCreated by {} on {}.".format(author, created_on)
                msg += "\nUpdated by {} on {}".format(updated_by, updated_on)

                if self.is_validator:
                    if callback.get("valid"):
                        msg += "\nThis callback is |rconnected|n and active."
                    else:
                        msg += "\nThis callback |rhasn't been|n accepted yet."

                msg += "\nCallback code:\n"
                msg += raw(callback["code"])
                self.msg(msg)
                return

            # No parameter has been specified, display the table of callbacks
            cols = ["Number", "Author", "Updated", "Param"]
            if self.is_validator:
                cols.append("Valid")

            table = EvTable(*cols, width=78)
            table.reformat_column(0, align="r")
            now = datetime.now()
            for i, callback in enumerate(created):
                author = callback.get("author")
                author = author.key if author else "|gUnknown|n"
                updated_on = callback.get("updated_on")
                if updated_on is None:
                    updated_on = callback.get("created_on")

                if updated_on:
                    updated_on = "{} ago".format(
                        time_format((now - updated_on).total_seconds(), 4).capitalize()
                    )
                else:
                    updated_on = "|gUnknown|n"
                parameters = callback.get("parameters", "")

                row = [str(i + 1), author, updated_on, parameters]
                if self.is_validator:
                    row.append("Yes" if callback.get("valid") else "No")
                table.add_row(*row)

            self.msg(str(table))
        else:
            names = list(set(list(types.keys()) + list(callbacks.keys())))
            table = EvTable("Callback name", "Number", "Description", valign="t", width=78)
            table.reformat_column(0, width=20)
            table.reformat_column(1, width=10, align="r")
            table.reformat_column(2, width=48)
            for name in sorted(names):
                number = len(callbacks.get(name, []))
                lines = sum(len(e["code"].splitlines()) for e in callbacks.get(name, []))
                no = "{} ({})".format(number, lines)
                description = types.get(name, (None, "Chained event."))[1]
                description = description.strip("\n").splitlines()[0]
                table.add_row(name, no, description)

            self.msg(str(table))

    def add_callback(self):
        """Add a callback."""
        obj = self.obj
        callback_name = self.callback_name
        types = self.handler.get_events(obj)

        # Check that the callback exists
        if not callback_name.startswith("chain_") and callback_name not in types:
            self.msg(
                "The callback name {} can't be found in {} of "
                "typeclass {}.".format(callback_name, obj, type(obj))
            )
            return

        definition = types.get(callback_name, (None, "Chained event."))
        description = definition[1]
        self.msg(raw(description.strip("\n")))

        # Open the editor
        callback = self.handler.add_callback(
            obj, callback_name, "", self.caller, False, parameters=self.parameters
        )

        # Lock this callback right away
        self.handler.db.locked.append((obj, callback_name, callback["number"]))

        # Open the editor for this callback
        self.caller.db._callback = callback
        EvEditor(
            self.caller,
            loadfunc=_ev_load,
            savefunc=_ev_save,
            quitfunc=_ev_quit,
            key="Callback {} of {}".format(callback_name, obj),
            persistent=True,
            codefunc=_ev_save,
        )

    def edit_callback(self):
        """Edit a callback."""
        obj = self.obj
        callback_name = self.callback_name
        parameters = self.parameters
        callbacks = self.handler.get_callbacks(obj)
        types = self.handler.get_events(obj)

        # If no callback name is specified, display the list of callbacks
        if not callback_name:
            self.list_callbacks()
            return

        # Check that the callback exists
        if callback_name not in callbacks:
            self.msg("The callback name {} can't be found in {}.".format(callback_name, obj))
            return

        # If there's only one callback, just edit it
        if len(callbacks[callback_name]) == 1:
            number = 0
            callback = callbacks[callback_name][0]
        else:
            if not parameters:
                self.msg("Which callback do you wish to edit?  Specify a number.")
                self.list_callbacks()
                return

            # Check that the parameter points to an existing callback
            try:
                number = int(parameters) - 1
                assert number >= 0
                callback = callbacks[callback_name][number]
            except (ValueError, AssertionError, IndexError):
                self.msg(
                    "The callback {} {} cannot be found in {}.".format(
                        callback_name, parameters, obj
                    )
                )
                return

        # If caller can't edit without validation, forbid editing
        # others' works
        if not self.autovalid and callback["author"] is not self.caller:
            self.msg("You cannot edit this callback created by someone else.")
            return

        # If the callback is locked (edited by someone else)
        if (obj, callback_name, number) in self.handler.db.locked:
            self.msg("This callback is locked, you cannot edit it.")
            return

        self.handler.db.locked.append((obj, callback_name, number))

        # Check the definition of the callback
        definition = types.get(callback_name, (None, "Chained event."))
        description = definition[1]
        self.msg(raw(description.strip("\n")))

        # Open the editor
        callback = dict(callback)
        self.caller.db._callback = callback
        EvEditor(
            self.caller,
            loadfunc=_ev_load,
            savefunc=_ev_save,
            quitfunc=_ev_quit,
            key="Callback {} of {}".format(callback_name, obj),
            persistent=True,
            codefunc=_ev_save,
        )

    def del_callback(self):
        """Delete a callback."""
        obj = self.obj
        callback_name = self.callback_name
        parameters = self.parameters
        callbacks = self.handler.get_callbacks(obj)
        types = self.handler.get_events(obj)

        # If no callback name is specified, display the list of callbacks
        if not callback_name:
            self.list_callbacks()
            return

        # Check that the callback exists
        if callback_name not in callbacks:
            self.msg("The callback name {} can't be found in {}.".format(callback_name, obj))
            return

        # If there's only one callback, just delete it
        if len(callbacks[callback_name]) == 1:
            number = 0
            callback = callbacks[callback_name][0]
        else:
            if not parameters:
                self.msg("Which callback do you wish to delete?  Specify " "a number.")
                self.list_callbacks()
                return

            # Check that the parameter points to an existing callback
            try:
                number = int(parameters) - 1
                assert number >= 0
                callback = callbacks[callback_name][number]
            except (ValueError, AssertionError, IndexError):
                self.msg(
                    "The callback {} {} cannot be found in {}.".format(
                        callback_name, parameters, obj
                    )
                )
                return

        # If caller can't edit without validation, forbid deleting
        # others' works
        if not self.autovalid and callback["author"] is not self.caller:
            self.msg("You cannot delete this callback created by someone else.")
            return

        # If the callback is locked (edited by someone else)
        if (obj, callback_name, number) in self.handler.db.locked:
            self.msg("This callback is locked, you cannot delete it.")
            return

        # Delete the callback
        self.handler.del_callback(obj, callback_name, number)
        self.msg("The callback {}[{}] of {} was deleted.".format(callback_name, number + 1, obj))

    def accept_callback(self):
        """Accept a callback."""
        obj = self.obj
        callback_name = self.callback_name
        parameters = self.parameters

        # If no object, display the list of callbacks to be checked
        if obj is None:
            table = EvTable("ID", "Type", "Object", "Name", "Updated by", "On", width=78)
            table.reformat_column(0, align="r")
            now = datetime.now()
            for obj, name, number in self.handler.db.to_valid:
                callbacks = self.handler.get_callbacks(obj).get(name)
                if callbacks is None:
                    continue

                try:
                    callback = callbacks[number]
                except IndexError:
                    continue

                type_name = obj.typeclass_path.split(".")[-1]
                by = callback.get("updated_by")
                by = by.key if by else "|gUnknown|n"
                updated_on = callback.get("updated_on")
                if updated_on is None:
                    updated_on = callback.get("created_on")

                if updated_on:
                    updated_on = "{} ago".format(
                        time_format((now - updated_on).total_seconds(), 4).capitalize()
                    )
                else:
                    updated_on = "|gUnknown|n"

                table.add_row(obj.id, type_name, obj, name, by, updated_on)
            self.msg(str(table))
            return

        # An object was specified
        callbacks = self.handler.get_callbacks(obj)
        types = self.handler.get_events(obj)

        # If no callback name is specified, display the list of callbacks
        if not callback_name:
            self.list_callbacks()
            return

        # Check that the callback exists
        if callback_name not in callbacks:
            self.msg("The callback name {} can't be found in {}.".format(callback_name, obj))
            return

        if not parameters:
            self.msg("Which callback do you wish to accept?  Specify a number.")
            self.list_callbacks()
            return

        # Check that the parameter points to an existing callback
        try:
            number = int(parameters) - 1
            assert number >= 0
            callback = callbacks[callback_name][number]
        except (ValueError, AssertionError, IndexError):
            self.msg(
                "The callback {} {} cannot be found in {}.".format(callback_name, parameters, obj)
            )
            return

        # Accept the callback
        if callback["valid"]:
            self.msg("This callback has already been accepted.")
        else:
            self.handler.accept_callback(obj, callback_name, number)
            self.msg(
                "The callback {} {} of {} has been accepted.".format(callback_name, parameters, obj)
            )

    def list_tasks(self):
        """List the active tasks."""
        obj = self.obj
        callback_name = self.callback_name
        handler = self.handler
        tasks = [(k, v[0], v[1], v[2]) for k, v in handler.db.tasks.items()]
        if obj:
            tasks = [task for task in tasks if task[2] is obj]
        if callback_name:
            tasks = [task for task in tasks if task[3] == callback_name]

        tasks.sort()
        table = EvTable("ID", "Object", "Callback", "In", width=78)
        table.reformat_column(0, align="r")
        now = datetime.now()
        for task_id, future, obj, callback_name in tasks:
            key = obj.get_display_name(self.caller)
            delta = time_format((future - now).total_seconds(), 1)
            table.add_row(task_id, key, callback_name, delta)

        self.msg(str(table))


# Private functions to handle editing


def _ev_load(caller):
    return caller.db._callback and caller.db._callback.get("code", "") or ""


def _ev_save(caller, buf):
    """Save and add the callback."""
    lock = "perm({}) or perm(events_without_validation)".format(WITHOUT_VALIDATION)
    autovalid = caller.locks.check_lockstring(caller, lock)
    callback = caller.db._callback
    handler = get_event_handler()
    if (
        not handler
        or not callback
        or not all(key in callback for key in ("obj", "name", "number", "valid"))
    ):
        caller.msg("Couldn't save this callback.")
        return False

    if (callback["obj"], callback["name"], callback["number"]) in handler.db.locked:
        handler.db.locked.remove((callback["obj"], callback["name"], callback["number"]))

    handler.edit_callback(
        callback["obj"], callback["name"], callback["number"], buf, caller, valid=autovalid
    )
    return True


def _ev_quit(caller):
    callback = caller.db._callback
    handler = get_event_handler()
    if (
        not handler
        or not callback
        or not all(key in callback for key in ("obj", "name", "number", "valid"))
    ):
        caller.msg("Couldn't save this callback.")
        return False

    if (callback["obj"], callback["name"], callback["number"]) in handler.db.locked:
        handler.db.locked.remove((callback["obj"], callback["name"], callback["number"]))

    del caller.db._callback
    caller.msg("Exited the code editor.")



# -*- coding: utf-8 -*-
# Generated by Django 1.9.9 on 2016-09-05 09:02


import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("accounts", "0004_auto_20150403_2339")]

    operations = [
        migrations.AlterField(
            model_name="accountdb",
            name="username",
            field=models.CharField(
                error_messages={"unique": "A user with that username already exists."},
                help_text="Required. 30 characters or fewer. Letters, digits and @/./+/-/_ only.",
                max_length=30,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        "^[\\w.@+-]+$",
                        "Enter a valid username. This value may contain only letters, numbers and @/./+/-/_ characters.",
                    )
                ],
                verbose_name="username",
            ),
        )
    ]



import datetime
import gzip
import pickle
import threading
from unittest import skipIf

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import ContentFile
from django.test import TestCase, override_settings
from django.utils.timezone import is_aware, utc

_SKIP = False
try:
    from botocore.exceptions import ClientError

    from .awsstorage import aws_s3_cdn as s3boto3
except ImportError:
    _SKIP = True


try:
    from django.utils.six.moves.urllib import parse as urlparse
except ImportError:
    from urllib import parse as urlparse


try:
    from unittest import mock
except ImportError:  # Python 3.2 and below
    import mock


@skipIf(_SKIP, "botocore not installed")
class S3Boto3TestCase(TestCase):
    def setUp(self):
        self.storage = s3boto3.S3Boto3Storage(access_key="foo", secret_key="bar")
        self.storage._connections.connection = mock.MagicMock()


@skipIf(_SKIP, "botocore not installed")
class S3Boto3StorageTests(S3Boto3TestCase):
    def test_clean_name(self):
        """
        Test the base case of _clean_name
        """
        path = self.storage._clean_name("path/to/somewhere")
        self.assertEqual(path, "path/to/somewhere")

    def test_clean_name_normalize(self):
        """
        Test the normalization of _clean_name
        """
        path = self.storage._clean_name("path/to/../somewhere")
        self.assertEqual(path, "path/somewhere")

    def test_clean_name_trailing_slash(self):
        """
        Test the _clean_name when the path has a trailing slash
        """
        path = self.storage._clean_name("path/to/somewhere/")
        self.assertEqual(path, "path/to/somewhere/")

    def test_clean_name_windows(self):
        """
        Test the _clean_name when the path has a trailing slash
        """
        path = self.storage._clean_name("path\\to\\somewhere")
        self.assertEqual(path, "path/to/somewhere")

    def test_pickle_with_bucket(self):
        """
        Test that the storage can be pickled with a bucket attached
        """
        # Ensure the bucket has been used
        self.storage.bucket
        self.assertIsNotNone(self.storage._bucket)

        # Can't pickle MagicMock, but you can't pickle a real Bucket object either
        p = pickle.dumps(self.storage)
        new_storage = pickle.loads(p)

        self.assertIsInstance(new_storage._connections, threading.local)
        # Put the mock connection back in
        new_storage._connections.connection = mock.MagicMock()

        self.assertIsNone(new_storage._bucket)
        new_storage.bucket
        self.assertIsNotNone(new_storage._bucket)

    def test_pickle_without_bucket(self):
        """
        Test that the storage can be pickled, without a bucket instance
        """

        # Can't pickle a threadlocal
        p = pickle.dumps(self.storage)
        new_storage = pickle.loads(p)

        self.assertIsInstance(new_storage._connections, threading.local)

    def test_storage_url_slashes(self):
        """
        Test URL generation.
        """
        self.storage.custom_domain = "example.com"

        # We expect no leading slashes in the path,
        # and trailing slashes should be preserved.
        self.assertEqual(self.storage.url(""), "https://example.com/")
        self.assertEqual(self.storage.url("path"), "https://example.com/path")
        self.assertEqual(self.storage.url("path/"), "https://example.com/path/")
        self.assertEqual(self.storage.url("path/1"), "https://example.com/path/1")
        self.assertEqual(self.storage.url("path/1/"), "https://example.com/path/1/")

    def test_storage_save(self):
        """
        Test saving a file
        """
        name = "test_storage_save.txt"
        content = ContentFile("new content")
        self.storage.save(name, content)
        self.storage.bucket.Object.assert_called_once_with(name)

        obj = self.storage.bucket.Object.return_value
        obj.upload_fileobj.assert_called_with(
            content,
            ExtraArgs={
                "ContentType": "text/plain",
                "ACL": self.storage.default_acl,
            },
        )

    def test_storage_save_with_acl(self):
        """
        Test saving a file with user defined ACL.
        """
        name = "test_storage_save.txt"
        content = ContentFile("new content")
        self.storage.default_acl = "private"
        self.storage.save(name, content)
        self.storage.bucket.Object.assert_called_once_with(name)

        obj = self.storage.bucket.Object.return_value
        obj.upload_fileobj.assert_called_with(
            content,
            ExtraArgs={
                "ContentType": "text/plain",
                "ACL": "private",
            },
        )

    def test_content_type(self):
        """
        Test saving a file with a None content type.
        """
        name = "test_image.jpg"
        content = ContentFile("data")
        content.content_type = None
        self.storage.save(name, content)
        self.storage.bucket.Object.assert_called_once_with(name)

        obj = self.storage.bucket.Object.return_value
        obj.upload_fileobj.assert_called_with(
            content,
            ExtraArgs={
                "ContentType": "image/jpeg",
                "ACL": self.storage.default_acl,
            },
        )

    def test_storage_save_gzipped(self):
        """
        Test saving a gzipped file
        """
        name = "test_storage_save.gz"
        content = ContentFile("I am gzip'd")
        self.storage.save(name, content)
        obj = self.storage.bucket.Object.return_value
        obj.upload_fileobj.assert_called_with(
            content,
            ExtraArgs={
                "ContentType": "application/octet-stream",
                "ContentEncoding": "gzip",
                "ACL": self.storage.default_acl,
            },
        )

    def test_storage_save_gzip(self):
        """
        Test saving a file with gzip enabled.
        """
        self.storage.gzip = True
        name = "test_storage_save.css"
        content = ContentFile("I should be gzip'd")
        self.storage.save(name, content)
        obj = self.storage.bucket.Object.return_value
        obj.upload_fileobj.assert_called_with(
            mock.ANY,
            ExtraArgs={
                "ContentType": "text/css",
                "ContentEncoding": "gzip",
                "ACL": self.storage.default_acl,
            },
        )
        args, kwargs = obj.upload_fileobj.call_args
        content = args[0]
        zfile = gzip.GzipFile(mode="rb", fileobj=content)
        self.assertEqual(zfile.read(), b"I should be gzip'd")

    def test_storage_save_gzip_twice(self):
        """
        Test saving the same file content twice with gzip enabled.
        """
        # Given
        self.storage.gzip = True
        name = "test_storage_save.css"
        content = ContentFile("I should be gzip'd")

        # When
        self.storage.save(name, content)
        self.storage.save("test_storage_save_2.css", content)

        # Then
        obj = self.storage.bucket.Object.return_value
        obj.upload_fileobj.assert_called_with(
            mock.ANY,
            ExtraArgs={
                "ContentType": "text/css",
                "ContentEncoding": "gzip",
                "ACL": self.storage.default_acl,
            },
        )
        args, kwargs = obj.upload_fileobj.call_args
        content = args[0]
        zfile = gzip.GzipFile(mode="rb", fileobj=content)
        self.assertEqual(zfile.read(), b"I should be gzip'd")

    def test_compress_content_len(self):
        """
        Test that file returned by _compress_content() is readable.
        """
        self.storage.gzip = True
        content = ContentFile("I should be gzip'd")
        content = self.storage._compress_content(content)
        self.assertTrue(len(content.read()) > 0)

    def test_storage_open_write(self):
        """
        Test opening a file in write mode
        """
        name = "test_open_for_writïng.txt"
        content = "new content"

        # Set the encryption flag used for multipart uploads
        self.storage.encryption = True
        self.storage.reduced_redundancy = True
        self.storage.default_acl = "public-read"

        file = self.storage.open(name, "w")
        self.storage.bucket.Object.assert_called_with(name)
        obj = self.storage.bucket.Object.return_value
        # Set the name of the mock object
        obj.key = name

        file.write(content)
        obj.initiate_multipart_upload.assert_called_with(
            ACL="public-read",
            ContentType="text/plain",
            ServerSideEncryption="AES256",
            StorageClass="REDUCED_REDUNDANCY",
        )

        # Save the internal file before closing
        multipart = obj.initiate_multipart_upload.return_value
        multipart.parts.all.return_value = [mock.MagicMock(e_tag="123", part_number=1)]
        file.close()
        multipart.Part.assert_called_with(1)
        part = multipart.Part.return_value
        part.upload.assert_called_with(Body=content.encode("utf-8"))
        multipart.complete.assert_called_once_with(
            MultipartUpload={"Parts": [{"ETag": "123", "PartNumber": 1}]}
        )

    def test_storage_open_no_write(self):
        """
        Test opening file in write mode and closing without writing.

        A file should be created as by obj.put(...).
        """
        name = "test_open_no_write.txt"

        # Set the encryption flag used for puts
        self.storage.encryption = True
        self.storage.reduced_redundancy = True
        self.storage.default_acl = "public-read"

        file = self.storage.open(name, "w")
        self.storage.bucket.Object.assert_called_with(name)
        obj = self.storage.bucket.Object.return_value
        obj.load.side_effect = ClientError(
            {"Error": {}, "ResponseMetadata": {"HTTPStatusCode": 404}}, "head_bucket"
        )

        # Set the name of the mock object
        obj.key = name

        # Save the internal file before closing
        file.close()

        obj.load.assert_called_once_with()
        obj.put.assert_called_once_with(
            ACL="public-read",
            Body=b"",
            ContentType="text/plain",
            ServerSideEncryption="AES256",
            StorageClass="REDUCED_REDUNDANCY",
        )

    def test_storage_open_no_overwrite_existing(self):
        """
        Test opening an existing file in write mode and closing without writing.
        """
        name = "test_open_no_overwrite_existing.txt"

        # Set the encryption flag used for puts
        self.storage.encryption = True
        self.storage.reduced_redundancy = True
        self.storage.default_acl = "public-read"

        file = self.storage.open(name, "w")
        self.storage.bucket.Object.assert_called_with(name)
        obj = self.storage.bucket.Object.return_value

        # Set the name of the mock object
        obj.key = name

        # Save the internal file before closing
        file.close()

        obj.load.assert_called_once_with()
        obj.put.assert_not_called()

    def test_storage_write_beyond_buffer_size(self):
        """
        Test writing content that exceeds the buffer size
        """
        name = "test_open_for_writïng_beyond_buffer_size.txt"

        # Set the encryption flag used for multipart uploads
        self.storage.encryption = True
        self.storage.reduced_redundancy = True
        self.storage.default_acl = "public-read"

        file = self.storage.open(name, "w")
        self.storage.bucket.Object.assert_called_with(name)
        obj = self.storage.bucket.Object.return_value
        # Set the name of the mock object
        obj.key = name

        # Initiate the multipart upload
        file.write("")
        obj.initiate_multipart_upload.assert_called_with(
            ACL="public-read",
            ContentType="text/plain",
            ServerSideEncryption="AES256",
            StorageClass="REDUCED_REDUNDANCY",
        )
        multipart = obj.initiate_multipart_upload.return_value

        # Write content at least twice as long as the buffer size
        written_content = ""
        counter = 1
        while len(written_content) < 2 * file.buffer_size:
            content = "hello, aws {counter}\n".format(counter=counter)
            # Write more than just a few bytes in each iteration to keep the
            # test reasonably fast
            content += "*" * int(file.buffer_size / 10)
            file.write(content)
            written_content += content
            counter += 1

        # Save the internal file before closing
        multipart.parts.all.return_value = [
            mock.MagicMock(e_tag="123", part_number=1),
            mock.MagicMock(e_tag="456", part_number=2),
        ]
        file.close()
        self.assertListEqual(multipart.Part.call_args_list, [mock.call(1), mock.call(2)])
        part = multipart.Part.return_value
        uploaded_content = "".join(
            args_list[1]["Body"].decode("utf-8") for args_list in part.upload.call_args_list
        )
        self.assertEqual(uploaded_content, written_content)
        multipart.complete.assert_called_once_with(
            MultipartUpload={
                "Parts": [
                    {"ETag": "123", "PartNumber": 1},
                    {"ETag": "456", "PartNumber": 2},
                ]
            }
        )

    def test_auto_creating_bucket(self):
        self.storage.auto_create_bucket = True
        Bucket = mock.MagicMock()
        self.storage._connections.connection.Bucket.return_value = Bucket
        self.storage._connections.connection.meta.client.meta.region_name = "sa-east-1"

        Bucket.meta.client.head_bucket.side_effect = ClientError(
            {"Error": {}, "ResponseMetadata": {"HTTPStatusCode": 404}}, "head_bucket"
        )
        self.storage._get_or_create_bucket("testbucketname")
        Bucket.create.assert_called_once_with(
            ACL="public-read",
            CreateBucketConfiguration={
                "LocationConstraint": "sa-east-1",
            },
        )

    def test_auto_creating_bucket_with_acl(self):
        self.storage.auto_create_bucket = True
        self.storage.bucket_acl = "public-read"
        Bucket = mock.MagicMock()
        self.storage._connections.connection.Bucket.return_value = Bucket
        self.storage._connections.connection.meta.client.meta.region_name = "sa-east-1"

        Bucket.meta.client.head_bucket.side_effect = ClientError(
            {"Error": {}, "ResponseMetadata": {"HTTPStatusCode": 404}}, "head_bucket"
        )
        self.storage._get_or_create_bucket("testbucketname")
        Bucket.create.assert_called_once_with(
            ACL="public-read",
            CreateBucketConfiguration={
                "LocationConstraint": "sa-east-1",
            },
        )

    def test_storage_exists(self):
        self.assertTrue(self.storage.exists("file.txt"))
        self.storage.connection.meta.client.head_object.assert_called_with(
            Bucket=self.storage.bucket_name,
            Key="file.txt",
        )

    def test_storage_exists_false(self):
        self.storage.connection.meta.client.head_object.side_effect = ClientError(
            {"Error": {"Code": "404", "Message": "Not Found"}},
            "HeadObject",
        )
        self.assertFalse(self.storage.exists("file.txt"))
        self.storage.connection.meta.client.head_object.assert_called_with(
            Bucket=self.storage.bucket_name,
            Key="file.txt",
        )

    def test_storage_exists_doesnt_create_bucket(self):
        with mock.patch.object(self.storage, "_get_or_create_bucket") as method:
            self.storage.exists("file.txt")
            self.assertFalse(method.called)

    def test_storage_delete(self):
        self.storage.delete("path/to/file.txt")
        self.storage.bucket.Object.assert_called_with("path/to/file.txt")
        self.storage.bucket.Object.return_value.delete.assert_called_with()

    def test_storage_listdir_base(self):
        # Files:
        #   some/path/1.txt
        #   2.txt
        #   other/path/3.txt
        #   4.txt
        pages = [
            {
                "CommonPrefixes": [
                    {"Prefix": "some"},
                    {"Prefix": "other"},
                ],
                "Contents": [
                    {"Key": "2.txt"},
                    {"Key": "4.txt"},
                ],
            },
        ]

        paginator = mock.MagicMock()
        paginator.paginate.return_value = pages
        self.storage._connections.connection.meta.client.get_paginator.return_value = paginator

        dirs, files = self.storage.listdir("")
        paginator.paginate.assert_called_with(Bucket=None, Delimiter="/", Prefix="")

        self.assertEqual(dirs, ["some", "other"])
        self.assertEqual(files, ["2.txt", "4.txt"])

    def test_storage_listdir_subdir(self):
        # Files:
        #   some/path/1.txt
        #   some/2.txt
        pages = [
            {
                "CommonPrefixes": [
                    {"Prefix": "some/path"},
                ],
                "Contents": [
                    {"Key": "some/2.txt"},
                ],
            },
        ]

        paginator = mock.MagicMock()
        paginator.paginate.return_value = pages
        self.storage._connections.connection.meta.client.get_paginator.return_value = paginator

        dirs, files = self.storage.listdir("some/")
        paginator.paginate.assert_called_with(Bucket=None, Delimiter="/", Prefix="some/")

        self.assertEqual(dirs, ["path"])
        self.assertEqual(files, ["2.txt"])

    def test_storage_size(self):
        obj = self.storage.bucket.Object.return_value
        obj.content_length = 4098

        name = "file.txt"
        self.assertEqual(self.storage.size(name), obj.content_length)

    def test_storage_mtime(self):
        # Test both USE_TZ cases
        for use_tz in (True, False):
            with self.settings(USE_TZ=use_tz):
                self._test_storage_mtime(use_tz)

    def _test_storage_mtime(self, use_tz):
        obj = self.storage.bucket.Object.return_value
        obj.last_modified = datetime.datetime.now(utc)

        name = "file.txt"
        self.assertFalse(
            is_aware(self.storage.modified_time(name)),
            "Naive datetime object expected from modified_time()",
        )

        self.assertIs(
            settings.USE_TZ,
            is_aware(self.storage.get_modified_time(name)),
            "{} datetime object expected from get_modified_time() when USE_TZ={}".format(
                ("Naive", "Aware")[settings.USE_TZ], settings.USE_TZ
            ),
        )

    def test_storage_url(self):
        name = "test_storage_size.txt"
        url = "http://aws.amazon.com/%s" % name
        self.storage.bucket.meta.client.generate_presigned_url.return_value = url
        self.storage.bucket.name = "bucket"
        self.assertEqual(self.storage.url(name), url)
        self.storage.bucket.meta.client.generate_presigned_url.assert_called_with(
            "get_object",
            Params={"Bucket": self.storage.bucket.name, "Key": name},
            ExpiresIn=self.storage.querystring_expire,
        )

        custom_expire = 123

        self.assertEqual(self.storage.url(name, expire=custom_expire), url)
        self.storage.bucket.meta.client.generate_presigned_url.assert_called_with(
            "get_object",
            Params={"Bucket": self.storage.bucket.name, "Key": name},
            ExpiresIn=custom_expire,
        )

    def test_generated_url_is_encoded(self):
        self.storage.custom_domain = "mock.cloudfront.net"
        filename = "whacky & filename.mp4"
        url = self.storage.url(filename)
        parsed_url = urlparse.urlparse(url)
        self.assertEqual(parsed_url.path, "/whacky%20%26%20filename.mp4")
        self.assertFalse(self.storage.bucket.meta.client.generate_presigned_url.called)

    def test_special_characters(self):
        self.storage.custom_domain = "mock.cloudfront.net"

        name = "ãlöhâ.jpg"
        content = ContentFile("new content")
        self.storage.save(name, content)
        self.storage.bucket.Object.assert_called_once_with(name)

        url = self.storage.url(name)
        parsed_url = urlparse.urlparse(url)
        self.assertEqual(parsed_url.path, "/%C3%A3l%C3%B6h%C3%A2.jpg")

    def test_strip_signing_parameters(self):
        expected = "http://bucket.s3-aws-region.amazonaws.com/foo/bar"
        self.assertEqual(
            self.storage._strip_signing_parameters(
                "%s?X-Amz-Date=12345678&X-Amz-Signature=Signature" % expected
            ),
            expected,
        )
        self.assertEqual(
            self.storage._strip_signing_parameters(
                "%s?expires=12345678&signature=Signature" % expected
            ),
            expected,
        )

    @skipIf(threading is None, "Test requires threading")
    def test_connection_threading(self):
        connections = []

        def thread_storage_connection():
            connections.append(self.storage.connection)

        for x in range(2):
            t = threading.Thread(target=thread_storage_connection)
            t.start()
            t.join()

        # Connection for each thread needs to be unique
        self.assertIsNot(connections[0], connections[1])

    def test_location_leading_slash(self):
        msg = (
            "S3Boto3Storage.location cannot begin with a leading slash. "
            "Found '/'. Use '' instead."
        )
        with self.assertRaises(ImproperlyConfigured, msg=msg):
            s3boto3.S3Boto3Storage(location="/")

    def test_override_class_variable(self):
        class MyStorage1(s3boto3.S3Boto3Storage):
            location = "foo1"

        storage = MyStorage1()
        self.assertEqual(storage.location, "foo1")

        class MyStorage2(s3boto3.S3Boto3Storage):
            location = "foo2"

        storage = MyStorage2()
        self.assertEqual(storage.location, "foo2")

    def test_override_init_argument(self):
        storage = s3boto3.S3Boto3Storage(location="foo1")
        self.assertEqual(storage.location, "foo1")
        storage = s3boto3.S3Boto3Storage(location="foo2")
        self.assertEqual(storage.location, "foo2")



# -*- coding: utf-8 -*-


from django.db import migrations, models


def convert_defaults(apps, schema_editor):
    ChannelDB = apps.get_model("comms", "ChannelDB")
    for channel in ChannelDB.objects.filter(db_typeclass_path="src.comms.comms.Channel"):
        channel.db_typeclass_path = "typeclasses.channels.Channel"
        channel.save()


class Migration(migrations.Migration):

    dependencies = [("comms", "0003_auto_20140917_0756")]

    operations = [migrations.RunPython(convert_defaults)]



"""
Module defining basic eventfuncs for the event system.

Eventfuncs are just Python functions that can be used inside of calllbacks.

"""

from evennia import ObjectDB
from evennia.contrib.base_systems.ingame_python.utils import InterruptEvent


def deny():
    """
    Deny, that is stop, the callback here.

    Notes:
        This function will raise an exception to terminate the callback
        in a controlled way.  If you use this function in an event called
        prior to a command, the command will be cancelled as well.  Good
        situations to use the `deny()` function are in events that begins
        by `can_`, because they usually can be cancelled as easily as that.

    """
    raise InterruptEvent


def get(**kwargs):
    """
    Return an object with the given search option or None if None is found.

    Keyword Args:
        Any searchable data or property (id, db_key, db_location...).

    Returns:
        The object found that meet these criteria for research, or
        None if none is found.

    Notes:
        This function is very useful to retrieve objects with a specific
        ID.  You know that room #32 exists, but you don't have it in
        the callback variables.  Quite simple:
            room = get(id=32)

        This function doesn't perform a search on objects, but a direct
        search in the database.  It's recommended to use it for objects
        you know exist, using their IDs or other unique attributes.
        Looking for objects by key is possible (use `db_key` as an
        argument) but remember several objects can share the same key.

    """
    try:
        object = ObjectDB.objects.get(**kwargs)
    except ObjectDB.DoesNotExist:
        object = None

    return object


def call_event(obj, event_name, seconds=0):
    """
    Call the specified event in X seconds.

    Args:
        obj (Object): the typeclassed object containing the event.
        event_name (str): the event name to be called.
        seconds (int or float): the number of seconds to wait before calling
                the event.

    Notes:
        This eventfunc can be used to call other events from inside of an
        event in a given time.  This will create a pause between events.  This
        will not freeze the game, and you can expect characters to move
        around (unless you prevent them from doing so).

        Variables that are accessible in your event using 'call()' will be
        kept and passed on to the event to call.

        Chained callbacks are designed for this very purpose: they
        are never called automatically by the game, rather, they need
        to be called from inside another event.

    """
    script = type(obj.callbacks).script
    if script:
        # If seconds is 0, call the event immediately
        if seconds == 0:
            locals = dict(script.ndb.current_locals)
            obj.callbacks.call(event_name, locals=locals)
        else:
            # Schedule the task
            script.set_task(seconds, obj, event_name)



"""
Turnbattle system - Tim Ashley Jenkins 2017

"""

from . import tb_basic  # noqa
from . import tb_equip  # noqa
from . import tb_items  # noqa
from . import tb_magic  # noqa
from . import tb_range  # noqa



# Generated by Django 3.2.3 on 2021-05-20 21:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("help", "0003_auto_20190128_1820"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="helpentry",
            name="db_staff_only",
        ),
        migrations.AddField(
            model_name="helpentry",
            name="db_date_created",
            field=models.DateTimeField(auto_now=True, verbose_name="creation date"),
        ),
    ]



"""
This structures the (simple) structure of the webpage 'application'.

"""
from django.urls import path

from . import views

app_name = "webclient"

urlpatterns = [path("", views.webclient, name="index")]



"""
The custom manager for Scripts.
"""

from django.conf import settings
from django.db.models import Q

from evennia.server import signals
from evennia.typeclasses.managers import TypeclassManager, TypedObjectManager
from evennia.utils.utils import class_from_module, dbid_to_obj, make_iter

__all__ = ("ScriptManager", "ScriptDBManager")
_GA = object.__getattribute__

_ObjectDB = None
_AccountDB = None


VALIDATE_ITERATION = 0


class ScriptDBManager(TypedObjectManager):
    """
    This Scriptmanager implements methods for searching
    and manipulating Scripts directly from the database.

    Evennia-specific search methods (will return Typeclasses or
    lists of Typeclasses, whereas Django-general methods will return
    Querysets or database objects).

    dbref (converter)
    dbref_search
    get_dbref_range
    object_totals
    typeclass_search
    get_all_scripts_on_obj
    get_all_scripts
    delete_script
    remove_non_persistent
    validate
    script_search (equivalent to evennia.search_script)
    copy_script

    """

    def get_all_scripts_on_obj(self, obj, key=None):
        """
        Find all Scripts related to a particular object.

        Args:
            obj (Object): Object whose Scripts we are looking for.
            key (str, optional): Script identifier - can be given as a
                dbref or name string. If given, only scripts matching the
                key on the object will be returned.
        Returns:
            matches (list): Matching scripts.

        """
        if not obj:
            return []
        account = _GA(_GA(obj, "__dbclass__"), "__name__") == "AccountDB"
        if key:
            dbref = self.dbref(key)
            if dbref or dbref == 0:
                if account:
                    return self.filter(db_account=obj, id=dbref)
                else:
                    return self.filter(db_obj=obj, id=dbref)
            elif account:
                return self.filter(db_account=obj, db_key=key)
            else:
                return self.filter(db_obj=obj, db_key=key)
        elif account:
            return self.filter(db_account=obj)
        else:
            return self.filter(db_obj=obj)

    def get_all_scripts(self, key=None):
        """
        Get all scripts in the database.

        Args:
            key (str or int, optional): Restrict result to only those
                with matching key or dbref.

        Returns:
            scripts (list): All scripts found, or those matching `key`.

        """
        if key:
            dbref = self.dbref(key)
            if dbref:
                return self.filter(id=dbref)
            return self.filter(db_key__iexact=key.strip())
        return self.all()

    def delete_script(self, dbref):
        """
        This stops and deletes a specific script directly from the
        script database.

        Args:
            dbref (int): Database unique id.

        Notes:
            This might be needed for global scripts not tied to a
            specific game object

        """
        scripts = self.get_id(dbref)
        for script in make_iter(scripts):
            script.stop()
            script.delete()

    def update_scripts_after_server_start(self):
        """
        Update/sync/restart/delete scripts after server shutdown/restart.

        """
        for script in self.filter(db_is_active=True, db_persistent=False):
            script._stop_task()

        for script in self.filter(db_is_active=True):
            script._unpause_task(auto_unpause=True)
            script.at_server_start()

        for script in self.filter(db_is_active=False):
            script.at_server_start()

    def search_script(self, ostring, obj=None, only_timed=False, typeclass=None):
        """
        Search for a particular script.

        Args:
            ostring (str): Search criterion - a script dbef or key.
            obj (Object, optional): Limit search to scripts defined on
                this object
            only_timed (bool): Limit search only to scripts that run
                on a timer.
            typeclass (class or str): Typeclass or path to typeclass.

        Returns:
            Queryset: An iterable with 0, 1 or more results.

        """

        ostring = ostring.strip()

        dbref = self.dbref(ostring)
        if dbref:
            # this is a dbref, try to find the script directly
            dbref_match = self.dbref_search(dbref)
            if dbref_match:
                dmatch = dbref_match[0]
                if not (obj and obj != dmatch.obj) or (only_timed and dmatch.interval):
                    return dbref_match

        if typeclass:
            if callable(typeclass):
                typeclass = "%s.%s" % (typeclass.__module__, typeclass.__name__)
            else:
                typeclass = "%s" % typeclass

        # not a dbref; normal search
        obj_restriction = obj and Q(db_obj=obj) or Q()
        timed_restriction = only_timed and Q(db_interval__gt=0) or Q()
        typeclass_restriction = typeclass and Q(db_typeclass_path=typeclass) or Q()
        scripts = self.filter(
            timed_restriction & obj_restriction & typeclass_restriction & Q(db_key__iexact=ostring)
        )
        return scripts

    # back-compatibility alias
    script_search = search_script

    def copy_script(self, original_script, new_key=None, new_obj=None, new_locks=None):
        """
        Make an identical copy of the original_script.

        Args:
            original_script (Script): The Script to copy.
            new_key (str, optional): Rename the copy.
            new_obj (Object, optional): Place copy on different Object.
            new_locks (str, optional): Give copy different locks from
                the original.

        Returns:
            script_copy (Script): A new Script instance, copied from
                the original.
        """
        typeclass = original_script.typeclass_path
        new_key = new_key if new_key is not None else original_script.key
        new_obj = new_obj if new_obj is not None else original_script.obj
        new_locks = new_locks if new_locks is not None else original_script.db_lock_storage

        from evennia.utils import create

        new_script = create.create_script(
            typeclass, key=new_key, obj=new_obj, locks=new_locks, autostart=True
        )
        return new_script

    def create_script(
        self,
        typeclass=None,
        key=None,
        obj=None,
        account=None,
        locks=None,
        interval=None,
        start_delay=None,
        repeats=None,
        persistent=None,
        autostart=True,
        report_to=None,
        desc=None,
        tags=None,
        attributes=None,
    ):
        """
        Create a new script. All scripts are a combination of a database
        object that communicates with the database, and an typeclass that
        'decorates' the database object into being different types of
        scripts.  It's behaviour is similar to the game objects except
        scripts has a time component and are more limited in scope.

        Keyword Args:
            typeclass (class or str): Class or python path to a typeclass.
            key (str): Name of the new object. If not set, a name of
                #dbref will be set.
            obj (Object): The entity on which this Script sits. If this
                is `None`, we are creating a "global" script.
            account (Account): The account on which this Script sits. It is
                exclusiv to `obj`.
            locks (str): one or more lockstrings, separated by semicolons.
            interval (int): The triggering interval for this Script, in
                seconds. If unset, the Script will not have a timing
                component.
            start_delay (bool): If `True`, will wait `interval` seconds
                before triggering the first time.
            repeats (int): The number of times to trigger before stopping.
                If unset, will repeat indefinitely.
            persistent (bool): If this Script survives a server shutdown
                or not (all Scripts will survive a reload).
            autostart (bool): If this Script will start immediately when
                created or if the `start` method must be called explicitly.
            report_to (Object): The object to return error messages to.
            desc (str): Optional description of script
            tags (list): List of tags or tuples (tag, category).
            attributes (list): List if tuples (key, value) or (key, value, category)
               (key, value, lockstring) or (key, value, lockstring, default_access).

        Returns:
            script (obj): An instance of the script created

        See evennia.scripts.manager for methods to manipulate existing
        scripts in the database.

        """
        global _ObjectDB, _AccountDB
        if not _ObjectDB:
            from evennia.accounts.models import AccountDB as _AccountDB
            from evennia.objects.models import ObjectDB as _ObjectDB

        typeclass = typeclass if typeclass else settings.BASE_SCRIPT_TYPECLASS

        if isinstance(typeclass, str):
            # a path is given. Load the actual typeclass
            typeclass = class_from_module(typeclass, settings.TYPECLASS_PATHS)

        # validate input
        kwarg = {}
        if key:
            kwarg["db_key"] = key
        if account:
            kwarg["db_account"] = dbid_to_obj(account, _AccountDB)
        if obj:
            kwarg["db_obj"] = dbid_to_obj(obj, _ObjectDB)
        if interval:
            kwarg["db_interval"] = max(0, interval)
        if start_delay:
            kwarg["db_start_delay"] = start_delay
        if repeats:
            kwarg["db_repeats"] = max(0, repeats)
        if persistent:
            kwarg["db_persistent"] = persistent
        if desc:
            kwarg["db_desc"] = desc
        tags = make_iter(tags) if tags is not None else None
        attributes = make_iter(attributes) if attributes is not None else None

        # create new instance
        new_script = typeclass(**kwarg)

        # store the call signature for the signal
        new_script._createdict = dict(
            key=key,
            obj=obj,
            account=account,
            locks=locks,
            interval=interval,
            start_delay=start_delay,
            repeats=repeats,
            persistent=persistent,
            autostart=autostart,
            report_to=report_to,
            desc=desc,
            tags=tags,
            attributes=attributes,
        )
        # this will trigger the save signal which in turn calls the
        # at_first_save hook on the typeclass, where the _createdict
        # can be used.
        new_script.save()

        if not new_script.id:
            # this happens in the case of having a repeating script with `repeats=1` and
            # `start_delay=False` - the script will run once and immediately stop before
            # save is over.
            return None

        signals.SIGNAL_SCRIPT_POST_CREATE.send(sender=new_script)

        return new_script


class ScriptManager(ScriptDBManager, TypeclassManager):
    pass
